/*
 * s5b.cpp - direct connection protocol via tcp
 * Copyright (C) 2003  Justin Karneges
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
  TODO:
    timeouts:
      between connects?
      after roleswitch?
    finish S5BConnection
      ByteStream stuff
      reset()
      progress signals
      proper close (if not accepted yet, send iq-error if possible)
      ensure deletable after signals
    save client()->jid() in connection, just in case it changes??
*/

#include"s5b.h"

#include<qtimer.h>
#include<stdlib.h>
#include<qca.h>
#include"xmpp_xmlcommon.h"
#include"socks.h"

namespace XMPP {

static QString makeKey(const QString &sid, const Jid &initiator, const Jid &target)
{
	QString str = sid + initiator.full() + target.full();
	QString result = QCA::SHA1::hashToString(str.utf8());
	printf("key=[%s]\n", result.latin1());
	return result;
}

//----------------------------------------------------------------------------
// S5BConnection
//----------------------------------------------------------------------------
class S5BConnection::Private
{
public:
	S5BManager *m;
	SocksClient *sc;
	int state;
	Jid peer;
	QString sid, key, iq_id;
	bool waiting;
	bool remote;
	bool prepared;
	bool switched;
	bool reuseme;
	JT_S5B *j;
	Jid streamHost;
	int id;
	S5BOutgoing *out;
	StreamHostList hosts;
};

static int id_conn = 0;
static int num_conn = 0;

S5BConnection::S5BConnection(S5BManager *m)
:ByteStream(m)
{
	d = new Private;
	d->m = m;
	d->sc = 0;
	d->j = 0;
	d->out = 0;

	++num_conn;
	d->id = id_conn++;
	printf("S5BConnection[%d]: constructing, count=%d\n", d->id, num_conn);

	reset();
}

S5BConnection::~S5BConnection()
{
	reset(true);

	--num_conn;
	printf("S5BConnection[%d]: destructing, count=%d\n", d->id, num_conn);

	delete d;
}

void S5BConnection::reset(bool)
{
	d->m->unlink(this);
	delete d->j;
	d->j = 0;
	delete d->out;
	d->out = 0;
	if(d->sc) {
		delete d->sc;
		d->sc = 0;
	}
	d->state = Idle;
	d->remote = false;
	d->waiting = false;
	d->prepared = false;
	d->switched = false;
	d->reuseme = false;
	d->iq_id = QString();
}

void S5BConnection::connectToJid(const Jid &peer, const QString &sid)
{
	reset(true);

	if(!d->m->isAcceptableSID(peer, sid))
		return;

	d->state = Requesting;
	d->peer = peer;
	d->sid = sid;
	d->m->link(this);

	doRequest();
}

void S5BConnection::doRequest()
{
	d->key = makeKey(d->sid, d->m->client()->jid(), d->peer);
	d->waiting = true; // we are waiting for an incoming socks connection
	printf("S5BConnection[%d]: initiating request %s [%s]\n", d->id, d->peer.full().latin1(), d->sid.latin1());

	d->j = new JT_S5B(d->m->client()->rootTask());
	connect(d->j, SIGNAL(finished()), SLOT(jt_finished()));

	StreamHostList hosts;
	S5BServer *serv = d->m->server();
	if(serv && serv->isActive()) {
		QStringList hostList = serv->hostList();
		for(QStringList::ConstIterator it = hostList.begin(); it != hostList.end(); ++it) {
			StreamHost h;
			h.setJid(d->m->client()->jid());
			h.setHost(*it);
			h.setPort(serv->port());
			hosts += h;
		}
	}

	d->j->request(d->peer, d->sid, hosts);
	d->iq_id = d->j->id();
	d->j->go(true);
}

void S5BConnection::accept()
{
	if(d->state != WaitingForAccept)
		return;

	d->state = Connecting;

	// connect in 60 seconds
	//d->t->start(60000, true);

	printf("S5BConnection[%d]: accepting %s [%s]\n", d->id, d->peer.full().latin1(), d->sid.latin1());

	d->out = new S5BOutgoing(d->m);
	connect(d->out, SIGNAL(result(bool)), SLOT(out_result(bool)));
	d->out->start(d->hosts, d->peer, d->key);
}

void S5BConnection::close()
{
	if(d->state == Idle)
		return;

	if(d->state == WaitingForAccept)
		d->m->doError(this, d->iq_id, 403, "Rejected");

	reset();
	printf("S5BConnection[%d]: closing %s [%s]\n", d->id, d->peer.full().latin1(), d->sid.latin1());
}

Jid S5BConnection::peer() const
{
	return d->peer;
}

QString S5BConnection::sid() const
{
	return d->sid;
}

bool S5BConnection::isRemote() const
{
	return d->remote;
}

int S5BConnection::state() const
{
	return d->state;
}

bool S5BConnection::isOpen() const
{
	if(d->state == Active)
		return true;
	else
		return false;
}

void S5BConnection::write(const QByteArray &buf)
{
	if(d->state == Active)
		d->sc->write(buf);
}

QByteArray S5BConnection::read(int bytes)
{
	if(d->sc)
		return d->sc->read(bytes);
	else
		return QByteArray();
}

int S5BConnection::bytesAvailable() const
{
	if(d->sc)
		return d->sc->bytesAvailable();
	else
		return 0;
}

int S5BConnection::bytesToWrite() const
{
	if(d->state == Active)
		return d->sc->bytesToWrite();
	else
		return 0;
}

void S5BConnection::jt_finished()
{
	if(d->state == Active)
		return;

	JT_S5B *j = d->j;
	d->j = 0;

	if(j->success()) {
		d->streamHost = j->streamHostUsed();

		if(!d->prepared) {
			// TODO
			printf("not prepared??\n");
			return;
		}

		d->state = Active;
		printf("S5BConnection[%d]: %s [%s] received successfully\n", d->id, d->peer.full().latin1(), d->sid.latin1());

		//d->t->stop();
		//QTimer::singleShot(0, this, SLOT(postConnect()));
		connected();

		// are we ready?
		/*if(!d->prepared) {
		}
		printf("S5BConnection[%d]: %s [%s] connected.\n", d->id, d->peer.full().latin1(), d->sid.latin1());

		d->state = Connected;
		connected();

		// connect in 60 seconds
		d->t->start(60000, true);

		// there might be waiting connections that need to be processed
		QTimer::singleShot(0, this, SLOT(postContinue()));
		d->m->continueAfterWait(d->keyA);*/
	}
	else {
		if(j->statusCode() == 300 && !d->switched) {
			printf("S5BConnection[%d]: %s [%s] role-switch!\n", d->id, d->peer.full().latin1(), d->sid.latin1());
			d->switched = true;
			d->reuseme = true;
		}
		else {
			printf("S5BConnection[%d]: %s [%s] refused.\n", d->id, d->peer.full().latin1(), d->sid.latin1());

			reset(true);
			error(ErrRequest);
		}
	}
}

QString S5BConnection::key() const
{
	return d->key;
}

bool S5BConnection::isWaiting() const
{
	return d->waiting;
}

void S5BConnection::setIncomingClient(SocksClient *sc)
{
	d->prepared = true;
	d->waiting = false;
	setSocksClient(sc);

	d->state = Active;
	printf("S5BConnection[%d]: %s [%s] received (connected) successfully\n", d->id, d->peer.full().latin1(), d->sid.latin1());
	connected();
}

void S5BConnection::setSocksClient(SocksClient *sc)
{
	d->sc = sc;
	connect(d->sc, SIGNAL(connectionClosed()), SLOT(sc_connectionClosed()));
	connect(d->sc, SIGNAL(delayedCloseFinished()), SLOT(sc_delayedCloseFinished()));
	connect(d->sc, SIGNAL(readyRead()), SLOT(sc_readyRead()));
	connect(d->sc, SIGNAL(bytesWritten(int)), SLOT(sc_bytesWritten(int)));
	connect(d->sc, SIGNAL(error(int)), SLOT(sc_error(int)));
}

void S5BConnection::waitForAccept(const Jid &peer, const QString &iq_id, const QString &sid, const StreamHostList &hosts)
{
	d->state = WaitingForAccept;
	d->remote = true;
	d->peer = peer;
	d->iq_id = iq_id;
	d->sid = sid;
	d->hosts = hosts;
	d->key = makeKey(sid, peer, d->m->client()->jid());

	d->m->link(this);
}

void S5BConnection::out_result(bool b)
{
	if(b) {
		d->state = Active;
		SocksClient *c = d->out->takeClient();
		StreamHost h = d->out->streamHostUsed();
		delete d->out;
		d->out = 0;

		setSocksClient(c);
		printf("S5BConnection[%d]: %s [%s] connected successfully.\n", d->id, d->peer.full().latin1(), d->sid.latin1());

		//d->t->stop();
		//QTimer::singleShot(0, this, SLOT(postConnect()));
		d->m->doSuccess(this, d->iq_id, h.jid());
		connected();
	}
	else {
		// try role-switch
		if(!d->switched) {
			d->switched = true;
			d->m->doError(this, d->iq_id, 300, "Role-switch");
			doRequest();
		}
		else {
			d->m->doError(this, d->iq_id, 502, "Could not connect to given hosts");
			reset(true);
			error(ErrConnect);
		}
	}
}

void S5BConnection::sc_connectionClosed()
{
	reset();
	connectionClosed();
}

void S5BConnection::sc_delayedCloseFinished()
{
	// echo
	delayedCloseFinished();
}

void S5BConnection::sc_readyRead()
{
	// echo
	readyRead();
}

void S5BConnection::sc_bytesWritten(int x)
{
	// echo
	bytesWritten(x);
}

void S5BConnection::sc_error(int)
{
	reset();
	error(ErrSocket);
}

//----------------------------------------------------------------------------
// S5BOutgoing
//----------------------------------------------------------------------------
class S5BOutgoing::Private
{
public:
	S5BManager *m;
	SocksClient *sc;
	StreamHostList hosts;
	Jid peer;
	QString key;
	StreamHost lastHost;
};

S5BOutgoing::S5BOutgoing(S5BManager *m)
:QObject(0)
{
	d = new Private;
	d->m = m;
	d->sc = 0;
}

S5BOutgoing::~S5BOutgoing()
{
	reset();
	delete d;
}

void S5BOutgoing::reset()
{
	delete d->sc;
	d->sc = 0;
	d->hosts.clear();
}

void S5BOutgoing::start(const StreamHostList &hosts, const Jid &peer, const QString &key)
{
	reset();

	d->hosts = hosts;
	d->peer = peer;
	d->key = key;

	QString dstr = "S5BOutgoing: trying ";
	bool first = true;
	StreamHostList::ConstIterator it;
	for(it = d->hosts.begin(); it != d->hosts.end(); ++it) {
		if(!first)
			dstr += ", ";
		dstr += (*it).host() + ':' + QString::number((*it).port());
		first = false;
	}
	dstr += '\n';
	printf("%s", dstr.latin1());

	QTimer::singleShot(0, this, SLOT(conn()));
}

SocksClient *S5BOutgoing::takeClient()
{
	SocksClient *c = d->sc;
	d->sc = 0;
	return c;
}

StreamHost S5BOutgoing::streamHostUsed() const
{
	return d->lastHost;
}

void S5BOutgoing::conn()
{
	if(d->hosts.isEmpty()) {
		result(false);
		return;
	}

	StreamHost h = d->hosts.first();
	d->lastHost = h;
	d->hosts.remove(d->hosts.begin());
	SocksClient *c = new SocksClient;
	connect(c, SIGNAL(connected()), SLOT(sc_connected()));
	connect(c, SIGNAL(error(int)), SLOT(sc_error(int)));
	d->sc = c;
	c->connectToHost(h.host(), h.port(), d->key, 0);
}

void S5BOutgoing::sc_connected()
{
	printf("S5BOutgoing: success\n");
	result(true);
}

void S5BOutgoing::sc_error(int)
{
	printf("S5BOutgoing: error\n");
	SocksClient *c = (SocksClient *)sender();
	c->deleteLater();
	d->sc = 0;

	QTimer::singleShot(0, this, SLOT(conn()));
}

//----------------------------------------------------------------------------
// S5BManager
//----------------------------------------------------------------------------
class S5BManager::Private
{
public:
	Client *client;
	S5BServer *serv;
	S5BConnectionList activeConns;
	S5BConnectionList incomingConns;
	JT_PushS5B *ps;
};

S5BManager::S5BManager(Client *parent)
:QObject(parent)
{
	d = new Private;
	d->client = parent;
	d->serv = 0;

	d->ps = new JT_PushS5B(d->client->rootTask());
	connect(d->ps, SIGNAL(incoming(const Jid &, const QString &, const QString &, const StreamHostList &)), SLOT(ps_incoming(const Jid &, const QString &, const QString &, const StreamHostList &)));
}

S5BManager::~S5BManager()
{
	setServer(0);
	d->incomingConns.setAutoDelete(true);
	d->incomingConns.clear();
	delete d->ps;
	delete d;
}

Client *S5BManager::client() const
{
	return d->client;
}

S5BServer *S5BManager::server() const
{
	return d->serv;
}

void S5BManager::setServer(S5BServer *serv)
{
	if(d->serv) {
		d->serv->unlink(this);
		d->serv = 0;
	}

	if(serv) {
		d->serv = serv;
		d->serv->link(this);
	}
}

S5BConnection *S5BManager::takeIncoming()
{
	if(d->incomingConns.isEmpty())
		return 0;

	S5BConnection *c = d->incomingConns.getFirst();
	d->incomingConns.removeRef(c);
	return c;
}

void S5BManager::ps_incoming(const Jid &from, const QString &id, const QString &sid, const StreamHostList &hosts)
{
	bool self = from.compare(d->client->jid());
	S5BConnection *c = findConnectionBySID(from, sid);
	bool loop = c && self && (id == c->d->iq_id);
	bool reuseme = c && c->d->reuseme;
	if(c && !loop && !reuseme) {
		d->ps->respondError(from, id, 403, "SID in use");
		return;
	}

	if(reuseme) {
		c->waitForAccept(from, id, sid, hosts);
		c->accept();
	}
	else {
		// create an incoming connection
		c = new S5BConnection(this);
		c->waitForAccept(from, id, sid, hosts);
		d->incomingConns.append(c);
		incomingReady();
	}
}

void S5BManager::doSuccess(S5BConnection *c, const QString &id, const Jid &streamHost)
{
	d->ps->respondSuccess(c->peer(), id, streamHost);
}

void S5BManager::doError(S5BConnection *c, const QString &id, int code, const QString &str)
{
	d->ps->respondError(c->peer(), id, code, str);
}

QString S5BManager::genUniqueSID(const Jid &peer) const
{
	// get unused key
	QString sid;
	do {
		sid = "s5b_";
		for(int i = 0; i < 4; ++i) {
			int word = rand() & 0xffff;
			for(int n = 0; n < 4; ++n) {
				QString s;
				s.sprintf("%x", (word >> (n * 4)) & 0xf);
				sid.append(s);
			}
		}
	} while(!isAcceptableSID(peer, sid));
	return sid;
}

bool S5BManager::isAcceptableSID(const Jid &peer, const QString &sid) const
{
	QString key = makeKey(sid, d->client->jid(), peer);

	// if we have a server, then check through it
	if(d->serv) {
		if(d->serv->findConnectionByHash(key))
			return false;
	}
	else {
		if(findConnectionByHash(key))
			return false;
	}
	return true;
}

void S5BManager::link(S5BConnection *c)
{
	d->activeConns.append(c);
}

void S5BManager::unlink(S5BConnection *c)
{
	d->activeConns.removeRef(c);
}

S5BConnection *S5BManager::findConnectionByHash(const QString &key, bool waitingOnly) const
{
	S5BConnectionListIt it(d->activeConns);
	for(S5BConnection *c; (c = it.current()); ++it) {
		if(!c->isWaiting() && waitingOnly)
			continue;
		if(c->key() == key)
			return c;
	}
	return 0;
}

S5BConnection *S5BManager::findConnectionBySID(const Jid &peer, const QString &sid) const
{
	S5BConnectionListIt it(d->activeConns);
	for(S5BConnection *c; (c = it.current()); ++it) {
		if(c->peer().compare(peer) && c->sid() == sid)
			return c;
	}
	return 0;
}

void S5BManager::ss_incomingReady(S5BConnection *c, SocksClient *sc)
{
	sc->requestGrant(true);
	c->setIncomingClient(sc);
}

//----------------------------------------------------------------------------
// S5BServer
//----------------------------------------------------------------------------
class S5BServer::Private
{
public:
	SocksServer serv;
	QStringList hostList;
	QPtrList<S5BManager> manList;
	QPtrList<SocksClient> clientList;
};

S5BServer::S5BServer(QObject *parent)
:QObject(parent)
{
	d = new Private;
	connect(&d->serv, SIGNAL(incomingReady()), SLOT(ss_incomingReady()));
}

S5BServer::~S5BServer()
{
	d->clientList.setAutoDelete(true);
	d->clientList.clear();
	delete d;
}

bool S5BServer::isActive() const
{
	return d->serv.isActive();
}

bool S5BServer::start(int port)
{
	d->serv.stop();
	return d->serv.listen(port);
}

void S5BServer::stop()
{
	d->serv.stop();
}

void S5BServer::setHostList(const QStringList &list)
{
	d->hostList = list;
}

QStringList S5BServer::hostList() const
{
	return d->hostList;
}

int S5BServer::port() const
{
	return d->serv.port();
}

void S5BServer::ss_incomingReady()
{
	SocksClient *c = d->serv.takeIncoming();
	connect(c, SIGNAL(incomingMethods(int)), SLOT(sc_incomingMethods(int)));
	connect(c, SIGNAL(incomingRequest(const QString &, int)), SLOT(sc_incomingRequest(const QString &, int)));
	connect(c, SIGNAL(error(int)), SLOT(sc_error(int)));
	d->clientList.append(c);
}

void S5BServer::sc_incomingMethods(int m)
{
	SocksClient *c = (SocksClient*)sender();
	if(m & SocksClient::AuthNone) {
		c->chooseMethod(SocksClient::AuthNone);
	}
	else {
		d->clientList.removeRef(c);
		c->disconnect(this);
		c->deleteLater();
	}
}

void S5BServer::sc_incomingRequest(const QString &host, int port)
{
	SocksClient *c = (SocksClient*)sender();
	d->clientList.removeRef(c);
	c->disconnect(this);

	if(port == 0) {
		S5BConnection *conn = findConnectionByHash(host, true);
		if(conn) {
			S5BManager *m = (S5BManager *)conn->parent();
			m->ss_incomingReady(conn, c);
			return;
		}
	}
	c->deleteLater();
}

void S5BServer::sc_error(int)
{
	SocksClient *c = (SocksClient*)sender();
	d->clientList.removeRef(c);
	c->disconnect(this);
	c->deleteLater();
}

void S5BServer::link(S5BManager *m)
{
	d->manList.append(m);
}

void S5BServer::unlink(S5BManager *m)
{
	d->manList.removeRef(m);
}

S5BConnection *S5BServer::findConnectionByHash(const QString &key, bool waitingOnly) const
{
	QPtrListIterator<S5BManager> it(d->manList);
	for(S5BManager *m; (m = it.current()); ++it) {
		S5BConnection *c = m->findConnectionByHash(key, waitingOnly);
		if(c)
			return c;
	}
	return 0;
}

S5BConnection *S5BServer::findConnectionBySID(const Jid &peer, const QString &sid) const
{
	QPtrListIterator<S5BManager> it(d->manList);
	for(S5BManager *m; (m = it.current()); ++it) {
		S5BConnection *c = m->findConnectionBySID(peer, sid);
		if(c)
			return c;
	}
	return 0;
}

//----------------------------------------------------------------------------
// JT_S5B
//----------------------------------------------------------------------------
class JT_S5B::Private
{
public:
	QDomElement iq;
	Jid to;
	Jid streamHost;
};

JT_S5B::JT_S5B(Task *parent)
:Task(parent)
{
	d = new Private;
}

JT_S5B::~JT_S5B()
{
	delete d;
}

void JT_S5B::request(const Jid &to, const QString &sid, const StreamHostList &hosts)
{
	QDomElement iq;
	d->to = to;
	iq = createIQ(doc(), "set", to.full(), id());
	QDomElement query = doc()->createElement("query");
	query.setAttribute("xmlns", "http://jabber.org/protocol/bytestreams");
	query.setAttribute("sid", sid);
	iq.appendChild(query);
	for(StreamHostList::ConstIterator it = hosts.begin(); it != hosts.end(); ++it) {
		QDomElement shost = doc()->createElement("streamhost");
		shost.setAttribute("jid", (*it).jid().full());
		shost.setAttribute("host", (*it).host());
		shost.setAttribute("port", QString::number((*it).port()));
		query.appendChild(shost);
	}
	d->iq = iq;
}

void JT_S5B::onGo()
{
	send(d->iq);
}

bool JT_S5B::take(const QDomElement &x)
{
	if(!iqVerify(x, d->to, id()))
		return false;

	if(x.attribute("type") == "result") {
		QDomElement q = queryTag(x);
		bool found;
		QDomElement shost = findSubTag(q, "streamhost-used", &found);
		if(found)
			d->streamHost = shost.attribute("jid");
		else
			d->streamHost = "";

		setSuccess();
	}
	else {
		setError(x);
	}

	return true;
}

Jid JT_S5B::streamHostUsed() const
{
	return d->streamHost;
}

//----------------------------------------------------------------------------
// JT_PushS5B
//----------------------------------------------------------------------------
JT_PushS5B::JT_PushS5B(Task *parent)
:Task(parent)
{
}

JT_PushS5B::~JT_PushS5B()
{
}

bool JT_PushS5B::take(const QDomElement &e)
{
	// must be an iq-set tag
	if(e.tagName() != "iq")
		return false;
	if(e.attribute("type") != "set")
		return false;
	if(queryNS(e) != "http://jabber.org/protocol/bytestreams")
		return false;

	Jid from(e.attribute("from"));
	QDomElement q = queryTag(e);
	QString sid = q.attribute("sid");

	StreamHostList hosts;
	QDomNodeList nl = q.elementsByTagName("streamhost");
	for(uint n = 0; n < nl.count(); ++n) {
		QDomElement shost = nl.item(n).toElement();
		if(hosts.count() < 3) {
			Jid j = shost.attribute("jid");
			if(!j.isValid())
				continue;
			StreamHost h;
			h.setJid(j);
			h.setHost(shost.attribute("host"));
			h.setPort(shost.attribute("port").toInt());
			hosts += h;
		}
	}

	incoming(from, e.attribute("id"), sid, hosts);
	return true;
}

void JT_PushS5B::respondSuccess(const Jid &to, const QString &id, const Jid &streamHost)
{
	QDomElement iq = createIQ(doc(), "result", to.full(), id);
	QDomElement query = doc()->createElement("query");
	query.setAttribute("xmlns", "http://jabber.org/protocol/bytestreams");
	iq.appendChild(query);
	QDomElement shost = doc()->createElement("streamhost-used");
	shost.setAttribute("jid", streamHost.full());
	query.appendChild(shost);
	send(iq);
}

void JT_PushS5B::respondError(const Jid &to, const QString &id, int code, const QString &str)
{
	QDomElement iq = createIQ(doc(), "error", to.full(), id);
	QDomElement err = textTag(doc(), "error", str);
	err.setAttribute("code", QString::number(code));
	iq.appendChild(err);
	send(iq);
}

//----------------------------------------------------------------------------
// StreamHost
//----------------------------------------------------------------------------
StreamHost::StreamHost()
{
}

const Jid & StreamHost::jid() const
{
	return j;
}

const QString & StreamHost::host() const
{
	return v_host;
}

int StreamHost::port() const
{
	return v_port;
}

void StreamHost::setJid(const Jid &_j)
{
	j = _j;
}

void StreamHost::setHost(const QString &host)
{
	v_host = host;
}

void StreamHost::setPort(int port)
{
	v_port = port;
}

}
