/*
 * s5b.h - direct connection protocol via tcp
 * Copyright (C) 2003  Justin Karneges
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef XMPP_S5B_H
#define XMPP_S5B_H

#include<qobject.h>
#include<qptrlist.h>
#include<qvaluelist.h>
#include"im.h"
#include"bytestream.h"

class SocksClient;

namespace XMPP
{
	class StreamHost;
	class S5BConnection;
	class S5BManager;
	class S5BServer;
	typedef QValueList<StreamHost> StreamHostList;
	typedef QPtrList<S5BConnection> S5BConnectionList;
	typedef QPtrListIterator<S5BConnection> S5BConnectionListIt;

	class S5BConnection : public ByteStream
	{
		Q_OBJECT
	public:
		enum { ErrRequest, ErrConnect, ErrSocket };
		enum { Idle, Requesting, Connecting, WaitingForAccept, Active };
		S5BConnection(S5BManager *);
		~S5BConnection();

		void connectToJid(const Jid &peer, const QString &sid);
		void accept();
		void close();

		Jid peer() const;
		QString sid() const;
		bool isRemote() const;
		int state() const;

		bool isOpen() const;
		void write(const QByteArray &);
		QByteArray read(int bytes=0);
		int bytesAvailable() const;
		int bytesToWrite() const;

	signals:
		void accepted();
		void connected();
		// void tryingStreamhost
		// void roleSwitching

	private slots:
		void sc_connectionClosed();
		void sc_delayedCloseFinished();
		void sc_readyRead();
		void sc_bytesWritten(int);
		void sc_error(int);

		void jt_finished();
		void out_result(bool);

		//void postAccept();
		//void postConnect();
		//void postContinue();
		//void t_timeout();

	private:
		class Private;
		Private *d;

		void reset(bool clear=false);
		void doRequest();

		friend class S5BManager;
		friend class S5BServer;
		void setIncomingClient(SocksClient *);
		void setSocksClient(SocksClient *);
		bool isWaiting() const;
		QString key() const;
		void waitForAccept(const Jid &peer, const QString &iq_id, const QString &sid, const StreamHostList &hosts);
	};

	class S5BOutgoing : public QObject
	{
		Q_OBJECT
	public:
		S5BOutgoing(S5BManager *);
		~S5BOutgoing();

		void reset();
		void start(const StreamHostList &hosts, const Jid &peer, const QString &key);
		SocksClient *takeClient();
		StreamHost streamHostUsed() const;

	signals:
		void result(bool);

	private slots:
		void sc_connected();
		void sc_error(int);
		void conn();

	private:
		class Private;
		Private *d;
	};

	class S5BManager : public QObject
	{
		Q_OBJECT
	public:
		S5BManager(Client *);
		~S5BManager();

		Client *client() const;
		S5BServer *server() const;
		void setServer(S5BServer *);
		bool isAcceptableSID(const Jid &peer, const QString &sid) const;
		QString genUniqueSID(const Jid &peer) const;

		S5BConnection *takeIncoming();

	signals:
		void incomingReady();

	private slots:
		void ps_incoming(const Jid &from, const QString &id, const QString &key, const StreamHostList &hosts);

	private:
		class Private;
		Private *d;

		friend class S5BConnection;
		friend class S5BServer;
		void link(S5BConnection *);
		void unlink(S5BConnection *);
		S5BConnection *findConnectionByHash(const QString &key, bool waitingOnly=false) const;
		S5BConnection *findConnectionBySID(const Jid &peer, const QString &sid) const;
		void ss_incomingReady(S5BConnection *c, SocksClient *sc);

		void doSuccess(S5BConnection *c, const QString &id, const Jid &streamHost);
		void doError(S5BConnection *c, const QString &id, int, const QString &);
	};

	// listens on a port for serving
	class S5BServer : public QObject
	{
		Q_OBJECT
	public:
		S5BServer(QObject *par=0);
		~S5BServer();

		bool isActive() const;
		bool start(int port);
		void stop();
		int port() const;
		void setHostList(const QStringList &);
		QStringList hostList() const;

	private slots:
		void ss_incomingReady();
		void sc_incomingMethods(int m);
		void sc_incomingRequest(const QString &user, int port);
		void sc_error(int);

	private:
		class Private;
		Private *d;

		friend class S5BManager;
		void link(S5BManager *);
		void unlink(S5BManager *);
		S5BConnection *findConnectionByHash(const QString &key, bool waitingOnly=false) const;
		S5BConnection *findConnectionBySID(const Jid &peer, const QString &sid) const;
	};

	class JT_S5B : public Task
	{
		Q_OBJECT
	public:
		JT_S5B(Task *);
		~JT_S5B();

		void request(const Jid &to, const QString &sid, const StreamHostList &hosts);

		void onGo();
		bool take(const QDomElement &);

		Jid streamHostUsed() const;

	private:
		class Private;
		Private *d;
	};

	class JT_PushS5B : public Task
	{
		Q_OBJECT
	public:
		JT_PushS5B(Task *);
		~JT_PushS5B();

		void respondSuccess(const Jid &to, const QString &id, const Jid &streamHost);
		void respondError(const Jid &to, const QString &id, int code, const QString &str);

		bool take(const QDomElement &);

	signals:
		void incoming(const Jid &from, const QString &id, const QString &sid, const StreamHostList &hosts);
	};

	class StreamHost
	{
	public:
		StreamHost();

		const Jid & jid() const;
		const QString & host() const;
		int port() const;
		void setJid(const Jid &);
		void setHost(const QString &);
		void setPort(int);

	private:
		Jid j;
		QString v_host;
		int v_port;
	};
}

#endif
